#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>

#include <stdio.h>

#include "etslib.h"

int debug = 0;

// Opens the modem and sets the baud rate
int tty_open(char *d, struct termios *saved_termios)
 {
   struct termios pts; 
   int fd, n;
   
   fd = open(d, O_RDWR|O_NDELAY);
   if (fd < 0)
     return -1;

   n = fcntl(fd, F_GETFL, 0);
   fcntl(fd, F_SETFL, n & ~O_NDELAY);

   if (tcgetattr(fd, saved_termios) != 0)
     {
       close(fd);
       return -1;
     }
   
   bzero(&pts, sizeof pts);
   pts.c_cflag = CRTSCTS | IGNPAR | CS8;
   pts.c_iflag = IGNPAR;
   pts.c_oflag = 0;
   pts.c_lflag = 0;

   pts.c_cc[VMIN] = 1;
   pts.c_cc[VTIME] = 0;

   cfsetospeed(&pts, B9600);
   cfsetispeed(&pts, B9600);

   tcflush(fd, TCIFLUSH);
   if (tcsetattr(fd, TCSANOW, &pts) != 0)
     {
       close(fd);
       return -1;
     }

   return fd;
 }

// Closes the modem device and resets the baudrate
void tty_close(int fd, struct termios *saved_termios)
{
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, saved_termios);
  close(fd);
}

int tty_write_byte(int fd, unsigned char b)
{
  if (debug)
    {
      if (b >= 32 && b < 127)
	fprintf(stderr, "WRITE: %c\n", b);
      else
	fprintf(stderr, "WRITE: 0x%02X\n", b);
    }
  
  return write(fd, &b, 1) == 1 ? 1 : 0;
}


#define TIMEOUT 5

int tty_read_byte(int fd)
{
  fd_set fds;
  int r;
  unsigned char c;
  struct timeval t;

  FD_ZERO(&fds);
  FD_SET(fd, &fds);

  t.tv_sec = 20;
  t.tv_usec = 0;

  r = select(FD_SETSIZE, &fds, 0, 0, &t);
  if (r != 1)
    return -1;

  if (read(fd, &c, 1) != 1)
    return -1;

  if (debug)
    {
      if (c >= 32 && c < 127)
	fprintf(stderr, " READ: %c\n", c);
      else
	fprintf(stderr, " READ: 0x%02X\n", c);
    }      

  return c;
}

char read_buffer[1024];
int read_buffer_pos = 0;
int read_buffer_len = 0;

char *tty_read_line(int fd)
{
  int quit = 0;
  int len = 256;
  char *p = malloc(len);
  char *e = p;
  
  while(!quit)
    {
      while (read_buffer_len > 0)
	{
	  char c = read_buffer[read_buffer_pos];

	  read_buffer_pos++;
	  read_buffer_len--;

	  *(e++) = c;
	      
	  if (c == '\n')
	    {
	      quit = 1;
	      break;
	    }
	  
	  if (e >= p+len-1)
	    {
	      quit = 1;
	      break;
	    }
	}

      if (quit) 
	break;

      read_buffer_pos = 0;
      read_buffer_len = read(fd, read_buffer, sizeof(read_buffer));

      if (read_buffer_len <= 0)
	{
	  read_buffer_len = 0;
	  break;
	}
    }
  
  *e = 0;

  return realloc(p, strlen(p)+1);
}

char *tty_chomp(char *p)
{
  char *e;

  while (1)
    {
      e = strchr(p, 0);
      if (e == p)
	break;

      if (*(e-1) != '\n' && *(e-1) != '\r')
	break;

      *(e-1) = 0;
    }
  
  return p;
}

void tty_flush(int fd)
{
  tcflush(fd, TCIFLUSH);
  tcflush(fd, TCIFLUSH);
}

int tty_can_read(int fd)
{
  fd_set fds;
  struct timeval t;

  FD_ZERO(&fds);
  FD_SET(fd, &fds);
  
  t.tv_sec = 0;
  t.tv_usec = 0;
  
  return (select(FD_SETSIZE, &fds, 0, 0, &t) == 1) ? 1: 0;
}


int ets_disconnect(int fd)
{
  tty_flush(fd);
  tty_write_byte(fd, 0xF0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xE0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xDF);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xCF);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xB0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xA0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  return 1;
}
     

char ets_identification[256];
int ets_identification_length = 0;

int ets_detect(int fd)
{
  int found;
  int i, b, l, c;
  tty_write_byte(fd, 0x18);
  tty_write_byte(fd, 0x18);


  found = 0;
  for (i = 0; i < 255; i++)
    {
      tty_write_byte(fd, '-');

      usleep(100000); // Zehntel Sekunde

      while (tty_can_read(fd) && !found)
	{
	  unsigned char r = tty_read_byte(fd);

	  if (r == 'A')
	    found = 1;
	}

      if (found) break;
    }
  
  tty_flush(fd);
  
  if (!found)
    return 0;
  
  tty_write_byte(fd, 0xF8);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xE0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xD0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  tty_flush(fd);
  tty_write_byte(fd, 0xC0);
  if (tty_read_byte(fd) != 'A')
    return 0;

  b = 1;
  l = -1;
  ets_identification_length = 0;
  while (1)
    {
      tty_flush(fd);
      if (b)
	tty_write_byte(fd, '?');
      else
	tty_write_byte(fd, '9');

      b = !b;

      c = tty_read_byte(fd);

      if (c < 0)
	return 0;

      if (c == '?' && l == '?')
	break;

      ets_identification[ets_identification_length++] = (unsigned char) c;

      l = c;
    }

  return ets_disconnect(fd);
}


/////////////// Datum lesen:
// Query
// * (f.Lnge-3) ! 0x00 0x00 0x00 0x00 0x00 (A1) (A2) (Datenlnge) (CHK1) (CHK2)
//               A
// Response
// (f.Lnge-3) " 0x00 0x00 0x02 0x00 0x00 (A1) (A2) (Datenlnge) (Daten1)...(DatenN) (CHK1) (CHK2)
//             B 

/////////////// Datum schreiben:
// Query
// * (f.Lnge-3) # 0x00 0x00 0x00 0x00 0x00 (A1) (A2) (Datenlnge) (Daten1)...(DatenN) (CHK1) (CHK2)
//
// Response
// (f.Lne-3) $ 0x00 0x00 0x02 0x00 0x00 (A1) (A2) 0x00 (CHK1) (CHK2)

// Adressen:
// Zeit:   0x84 0x99 / 0x02 (!)
//         0x84 0x9B / 0x06 (!)  **Hierein wird auch geschrieben**
//         0x00 0x9D / 0x02 (A)
//         0x00 0x9D / 0x02 (A)

// Konfig: 0x80 0x1A / 0x2B (!)
//         0x80 0x55 / 0x10 (!)
//         0x80 0x75 / 0xF3 (!)
//         0x81 0x68 / 0xF3 (!)
//         0x82 0x5B / 0xF3 (!)
//         0x83 0x4E / 0x43 (!)
//         0x83 0x95 / 0xF3 (!)
//         0x84 0x88 / 0x11 (!)
//         0x84 0xAB / 0xF3 (!)
//         0x85 0x9E / 0xB2 (!)
//         0x86 0x53 / 0xF3 (!)
//         0x87 0x46 / 0xF3 (!)
//         0x88 0x39 / 0xF3 (!)
//         0x89 0x2C / 0xF3 (!)
//         0x8A 0x1F / 0xF3 (!)
//         0x8B 0x02 / 0xF3 (!)
//         0x8C 0x05 / 0xF3 (!)
//         0x8C 0xF8 / 0xF3 (!)
//         0x8D 0xEB / 0xF3 (!)
//         0x8E 0xDE / 0xF3 (!)
//         0x8F 0xD1 / 0xF3 (!)
//         0x90 0xC4 / 0xA7 (!)
//         0x92 0x0B / 0xF3 (!)
//         0x92 0xFE / 0x6B (!)
//         0xA5 0x35 / 0xF3 (!)
//         0xA6 0x28 / 0x4D (!)
//         0xA6 0xA2 / 0x01 (!)
//         0xA6 0xA5 / 0x06 (!)
//         0xA6 0xE9 / 0x10 (!)


int ets_get_time(int fd)
{
  
}
