//
// ETSRE - ETS 4308i reverse Engeneering Program
// Copyright (C) 2000 Lennart Poettering <lennart@poettering.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//			
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//					
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	    
//

#include <termios.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <sys/stat.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/time.h>


// Where do lockfiles reside?
#define LOCK_PATH "/var/lock"

// Baudrate for the TTY.
#define BAUD_RATE B9600

// Modem-device
char *device = 0;

// Our name
char *appname = 0;

// The name of our user
char *username = 0;

// Enable debug-mode or not?
int debug = 0;

// We got a ctrl-c
volatile int interrupted = 0;

// Shall we wait for a call?
int answer = 0;

// After how many RINGs we shall answer?
int rings = 1;

// Waits for a certain string from the modem up to a certain time.
int tty_waitfor(int fd, char *what, int to)
{
  char *p = what;
  int quit = 0;

  if (debug)
    fprintf(stderr, "<- ");

  while (! quit)
    {
      int t, i;
      unsigned char c;
      fd_set fds;
      struct timeval tv, *tvp;

      if (interrupted)
	break;

      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      if (to != -1)
	{
	  tv.tv_sec = to;
	  tv.tv_usec = 0;
      
	  tvp = &tv;
	}
      else 
	tvp = 0;


      if (select(FD_SETSIZE, &fds, 0, 0, tvp) == 1)
	{
	  t = read(fd, &c, 1);

	  if (t == 1)
	    {
	      if (debug)
		fprintf(stderr, "%c", c >= 32 && c <= 126 ? c : '.');
	      
	      for (i = 0; i < t; i++)
		{
		  if (c == *p)
		    {
		      p++;
		      if (*p == 0)
			{
			  quit = 1;
			  break;
			}
		    }
		  else
		    {
		      p = what;
		    }
		  
		}
	    }
	}
      else
	break;
    }
  
  if (debug)
    fprintf(stderr, "\n");
  
  return quit;
}

// Write a string to the modem
void tty_puts(int fd, char *s)
 {
   write (fd, s, strlen(s));
  
   if (debug) 
     fprintf(stderr, "-> %s", s);
 }

// The name says it all.
void flush_available_data(int fd)
{
  tcflush(fd, TCIFLUSH);
  tcflush(fd, TCIFLUSH);
}


// Issue a command to the modem an wait for an answer
int tty_command(int fd, char *cmd, char* resp, int t)
{
  int r;
  flush_available_data(fd);

  tty_puts(fd, cmd);

  r = tty_waitfor(fd, resp, t);
  
  return r;
}

// Hang up the modem
void hard_hangup(int fd)
{
   struct termios pts;

   tcgetattr(fd, &pts);
   cfsetospeed(&pts, B0);
   cfsetispeed(&pts, B0);
   tcsetattr(fd, TCSANOW, &pts);

   sleep(1);

   tcgetattr(fd, &pts);
   cfsetospeed(&pts, BAUD_RATE);
   cfsetispeed(&pts, BAUD_RATE);
   tcsetattr(fd, TCSANOW, &pts);

   tty_puts(fd, "AT\n");

   sleep(1);
   
   flush_available_data(fd);
}

struct termios saved_termios;


// Opens the modem and sets the baud rate
int tty_open()
 {
   struct termios pts; 
   int fd, n;
   
   fd = open(device, O_RDWR|O_NDELAY);
   if (fd < 0)
     return -1;

   n = fcntl(fd, F_GETFL, 0);
   fcntl(fd, F_SETFL, n & ~O_NDELAY);

   if (tcgetattr(fd, &saved_termios) != 0)
     {
       close(fd);
       return -1;
     }
   
   bzero(&pts, sizeof pts);
   pts.c_cflag = CRTSCTS | IGNPAR | HUPCL | CS8;
   pts.c_iflag = IGNPAR;
   pts.c_oflag = 0;
   pts.c_lflag = 0;

   pts.c_cc[VMIN] = 1;
   pts.c_cc[VTIME] = 0;

   cfsetospeed(&pts, BAUD_RATE);
   cfsetispeed(&pts, BAUD_RATE);

   tcflush(fd, TCIFLUSH);
   if (tcsetattr(fd, TCSANOW, &pts) != 0)
     {
       close(fd);
       return -1;
     }

   return fd;
 }

// Closes the modem device and resets the baudrate
void tty_close(int fd)
{
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, &saved_termios);
  close(fd);
}

// Detects the DCD line
int detect_carrier(int fd)
{
  unsigned int arg;
  int status = ioctl(fd, TIOCMGET, &arg);
  if (status == 0)
    return !! (arg & TIOCM_CAR);
  else
    return -1;
}

char lockfile[PATH_MAX] = "";

// Returns a pointer to the filename in a path
char *basename(char *f)
{
  char *i = strrchr(f, '/');

  return i ? i+1 : f;
}

// Tries to lock a TTY-device
int tty_lock(char *dev)
{
  struct stat stt;
  char buf[256];

  if (stat(LOCK_PATH, &stt) == 0)
    {
      int fd;
      snprintf(lockfile, sizeof(lockfile), "%s/LCK..%s", LOCK_PATH, basename(dev));
      
      if ((fd = open(lockfile, O_RDONLY)) >= 0) 
	{
	  int n = read(fd, buf, sizeof(buf) - 1);
	  close(fd);

	  if (n > 0)
	    {
	      int pid;

	      if (n == 4)
		{
		  pid = *(int*) buf;
		}
	      else
		{
		  buf[n] = 0;
		  sscanf(buf, "%d", &pid);
		}
	      
	      if (pid > 0)
		{
		  if (kill((pid_t) pid, 0) < 0 && errno == ESRCH) 
		    {
		      fprintf(stderr, "Lockfile is stale. Overriding it.\n");
		      unlink(lockfile);
		    }
		  else
		    return 0;
		}
	    }
	}

      if ((fd = open(lockfile, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0)
	{
	  return 0;
	}
      
      snprintf(buf, sizeof(buf), "%10ld %s %.20s\n", (long)getpid(), appname, username);
      write(fd, buf, strlen(buf));
      close(fd);

      return 1;
    }
  else
    return 0;
}

// Tries to lock the first free device
int tty_lock_first_free_device()
{
  int i, found = 0;
  char dev[256];

  for (i = 2; i <= 20; i++)
    {
      snprintf(dev, sizeof(dev), "/dev/ttyI%i", i);

      if (tty_lock(dev))
	{
	  found = 1;
	  break;
	}
    }

  if (found)
    {
      if (debug)
	fprintf(stderr, "Using device %s.\n", dev);

      if (device) free(device);
      device = strdup (dev);
    }
  else
    if (debug)
      fprintf(stderr, "Could not find free device.\n");

  return found;
}


// Deletes the lockfile
void tty_unlock()
{
  unlink(lockfile);
}

// Filters out DLE-sequences from modem
char *filter_dle(unsigned char *data, int *size, int *flag, int *quit)
{
  char *ret;
  int sidx, didx;

  ret = malloc(*size);
  didx = 0;

  for (sidx = 0; sidx < *size; sidx++)
    {
      if (! *flag)
	{
	  if (data[sidx] == DLE)
	    *flag = 1;
	  else
	    ret[didx++] = data[sidx];
	}
      else
	{
	  switch (data[sidx])
	    {
	    case DLE :
	      ret[didx++] = DLE;		  
	      break;
		
	    case ETX :
	    case DC4 :
	      *quit = 1;
	      break;
	      
	    default :
	      // We recieved some DTMF signals, but we ignore them.
	      break;
	    }

	  *flag = 0;
	}


    }


  *size = didx;

  return realloc(ret, *size);
}

// Escape DLE-bytes for sending data to the modem
char* escape_dle(unsigned char *data, int*size)
{
  unsigned char *ret;
  int sidx, didx, cs;

  ret = malloc((int)(*size*1.05));
  cs = *size;
  didx = 0;

  for (sidx = 0; sidx < *size; sidx++)
    {
      if (data[sidx] == DLE)
	{
	  ret = realloc(ret, ++cs);
	  ret[didx++] = DLE;
	}

      ret[didx++] = data[sidx];
    }

  *size = cs;
  return realloc(ret, *size);
}

// Our ctrl-c handler
void sigint()
{
  interrupted = 1;
}
 
// The function that does the modem work
int go()
{
  int fd;
  char buf[1024];
  int i;
  int flag;
  int quit;
  int ret = 0, warned = 0;
  int lv;

  signal(SIGINT, &sigint);

  if (device)
    lv = tty_lock(device);
  else
    lv = tty_lock_first_free_device();


  if (! lv)
    {
      fprintf(stderr, "Could not get lock on device.\n");
      ret = 1;
    }
  else
    {
      fd = tty_open();
      if (fd < 0)
	{
	  fprintf(stderr, "Could open device.\n");
	  ret = 1;
	}
      else
	{
	  for (i = 0; at_commands[i] != 0; i+=2)
	    {
	      if (! tty_command(fd, at_commands[i], at_commands[i+1], COMMAND_TIMEOUT))
		{
		  fprintf(stderr, "Initialisation failure.\n");
		  ret = 1;
		  break;
		}
	    }    
	  
	  if (!ret)
	    {
	      snprintf(buf, sizeof(buf), "AT&E%s\n", source_msn);
	      if (! tty_command(fd, buf, "OK\r\n", COMMAND_TIMEOUT))
		{
		  fprintf(stderr, "Initialisation failure.\n");
		  ret = 1;
		}
	      else
		{
		  if (answer)
		    snprintf(buf, sizeof(buf), "ATS0=%i\n", rings); 
		  else
		    snprintf(buf, sizeof(buf), "ATD%s\n", destination_msn); 

		  if (! tty_command(fd, buf, "VCON\r\n", timeout))
		    {
		      ret = 2;
		    }
		  else
		    {
		      
		      flush_available_data(fd);
		      
		      if (! tty_command(fd, "AT+VTX+VRX\n", "CONNECT\r\n\r\nCONNECT\r\n", COMMAND_TIMEOUT))
			{
			  fprintf(stderr, "Could not switch to full-duplex audio.\n");
			  ret = 1;
			}
		      else
			{
			  flag = quit = ret = 0;
			  
			  while (!(ret || quit))
			    {
			      fd_set fds;
			      int sr;
			      int carrier = detect_carrier(fd);
			      
			      if (carrier <= 0)
				{
				  quit = 1;
				  continue;
				}

			      FD_ZERO(&fds);
			      FD_SET(0, &fds);
			      FD_SET(fd, &fds);
			      
			      sr = select(FD_SETSIZE, &fds, 0, 0, 0);

			      if (interrupted)
				{
				  fprintf(stderr, "Exiting because of user break.\n");
				  ret = 1;
				  continue;
				}

			      if (sr <= 0)
				{
				  fprintf(stderr, "select() returned error.\n");
				  ret = 1;
				  continue;
				}
				  
			      if (FD_ISSET(0, &fds))
				{
				  int l = read(0, buf, sizeof(buf));
				  
				  if (l <= 0) 
				    {
				      quit = 1;
				    }
				  else
				    {
				      int size = l;
				      char *nbuf = escape_dle(buf, &size);
				      write (fd, nbuf, size);		
				      free(nbuf);
				    }
				}
			      
			      if (FD_ISSET(fd, &fds))
				{
				  int l;
				  
				  l = read(fd, buf, sizeof(buf));
				  if (l <= 0)
				    {
				      quit = 1;
				    }
				  else
				    {
				      int size = l;
				      char *nbuf = filter_dle(buf, &size, &flag, &quit);

				      if (isatty(1) && !warned)
					{
					  fprintf(stderr, "I won't print out binary data to a tty. Pipe stdout to somewhere else.\n");
					  warned = 1;
					}
				      
				      if (!isatty(1))
					write (1, nbuf, size);
				      free(nbuf);
				    }
				}
			    }
			  
			  buf[0] = DLE;
			  buf[1] = DC4;
			  write (fd, buf, 2);
			  
			  buf[0] = DLE;
			  buf[1] = ETX;
			  write (fd, buf, 2);
			}
		      
		      tty_puts(fd, "ATH\n");
		      
		      hard_hangup(fd);
		    }

		}
		  
	      tty_puts(fd, "AT&F\n");
	      
	      sleep(1);
	      
	      flush_available_data(fd);
	      flush_available_data(0);
	      
	    }
	  tty_close(fd);
	}
      tty_unlock();
    }
  
  signal(SIGINT, SIG_DFL);
  
  return ret;
}

struct option long_options[] =
 {
  {"help", 0, 0, 'h'},
  {"timeout", 1, 0, 't'},
  {"device", 1, 0, 'd'},
  {"verbose", 0, 0, 'v'},
  {"answer", 0, 0, 'a'},
  {"ring", 1, 0, 'r'},
  {0, 0, 0, 0}
 };

// Most important part of the program
void usage()
{
  printf("Usage: %s [options] [source-msn] [destination-MSN]\n", appname);
  printf("\n");
  printf("Options:\n");
  printf(" -h, --help            shows this usage information.\n");
  printf(" -t, --timeout=TIMEOUT sets a timeout.\n");
  printf(" -d, --device=DEVICE   selects a device.\n");
  printf("                       (If not specified, the first free will be taken.\n");
  printf(" -v, --verbose         enables debug mode.\n");
  printf(" -a, --answer          wait for incoming call instead of calling out.\n");
  printf(" -r, --ring=RINGS      on incoming call answer after given RINGs.\n");
}

// Parses command line and executes go()
int main (int argc, char *argv[])
{
  int quit = 0;
  int help = 0;
  int ret;
  struct passwd *pwd;

  pwd = getpwuid(getuid());
  if (!pwd)
    {
      fprintf(stderr, "You don't exist. Go away.\n");
      return -1;
    }
  username = strdup(pwd->pw_name);

  appname = strdup(basename(argv[0]));

  while (1)
    {
      int option_index;
      int c;

      c = getopt_long(argc, argv, "ht:d:var:", long_options, &option_index);
      if (c == -1)
	break;
      
      switch (c)
	{
	case 'h' :
	  help = 1;
	  break;
	  
	case 't' :
	  if (optarg)
	    {
	      timeout = atoi(optarg);
	      if (timeout < 1)
		{
		  fprintf(stderr, "Timeout must be >= 1.\n");
		  quit = 1;
		}
	    }
	  else
	    {
	      fprintf(stderr, "Expected argument on timeout option.\n");
	      quit = 1;
	    }
	  break;

	case 'd' :
	  if (optarg)
	    {
	      if (device) 
		free(device);
	      else
		device = strdup(optarg);
	    }
	  else
	    {
	      fprintf(stderr, "Expected argument on device option.\n");
	      quit = 1;
	    }
	  break;

	case 'v' :
	  debug = 1;
	  break;
	  
	case 'a' :
	  answer = 1;
	  break;
	  
	case 'r' :
	  if (optarg)
	    {
	      rings = atoi(optarg);
	      if (rings < 1)
		{
		  fprintf(stderr, "RINGs must be >= 1.\n");
		  quit = 1;
		}
	    }
	  else
	    {
	      fprintf(stderr, "Expected argument on rings option.\n");
	      quit = 1;
	    }
	  break;

	case '?' :
	  quit = 1;
	  break;
	}

      if (quit) 
	break;
    }
  
  if (quit)
    {
      if (device) free (device);
      if (username) free(username);      
      if (appname) free(appname);
      return 1;
    }

  if (!help)
    {
      if (answer)
	{
	  if (optind != argc-1)
	    {
	      fprintf(stderr, "You did not specify the listening MSN.\n");
	      if (optind == argc) 
		help = 1;
	      else 
		quit = 1;
	    }
	  else
	    source_msn = strdup(argv[optind]);
	}
      else
	{
	  if (optind != argc-2)
	    {
	      fprintf(stderr, "You did not specify both destination and source MSNs.\n");
	      if (optind == argc) 
		help = 1;
	      else
		quit = 1;
	    }
	  else
	    {
	      source_msn = strdup(argv[optind]);
	      destination_msn = strdup(argv[optind+1]);
	    }
	}
    }
  
  if (help) 
    usage();

  if (help || quit)
    {
      if (username) free(username);      
      if (appname) free(appname);
      if (source_msn) free (source_msn);
      if (destination_msn) free(destination_msn);
      if (device) free(device);

      return help ? 0 : 2;
    }
    

  if (timeout == 0) 
    {
      if (answer)
	timeout = -1;
      else
	timeout = DEFAULT_TIMEOUT;
    }      

  ret = go();

  if (username) free(username);
  if (appname) free(appname);
  if (source_msn) free (source_msn);
  if (destination_msn) free(destination_msn);
  if (device) free(device);
    
  return ret;
}

